@@ -29,6 +29,9 @@
 
 #include "connman.h"
 
+#define MAX_CONNECT_RETRIES	2
+#define CONNECT_TIMEOUT		120
+
 static DBusConnection *connection = NULL;
 
 static GSequence *service_list = NULL;
@@ -47,11 +50,20 @@ struct connman_service {
 	connman_bool_t favorite;
 	connman_bool_t hidden;
 	connman_bool_t ignore;
+	connman_bool_t autoconnect;
+	connman_bool_t userconnect;
 	GTimeVal modified;
 	unsigned int order;
+	unsigned int failcounter;
 	char *name;
 	char *passphrase;
 	char *profile;
+	char *apn;
+	char *username;
+	char *password;
+	char *mcc;
+	char *mnc;
+	connman_bool_t roaming;
 	struct connman_ipconfig *ipconfig;
 	struct connman_device *device;
 	struct connman_network *network;
@@ -135,6 +147,12 @@ static const char *mode2string(enum connman_service_mode mode)
 		return "managed";
 	case CONNMAN_SERVICE_MODE_ADHOC:
 		return "adhoc";
+	case CONNMAN_SERVICE_MODE_GPRS:
+		return "gprs";
+	case CONNMAN_SERVICE_MODE_EDGE:
+		return "edge";
+	case CONNMAN_SERVICE_MODE_UMTS:
+		return "umts";
 	}
 
 	return NULL;
@@ -230,6 +248,36 @@ const char *__connman_service_default(void)
 	return type2string(service->type);
 }
 
+static void mode_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *str, *key = "Mode";
+
+	if (service->path == NULL)
+		return;
+
+	str = mode2string(service->mode);
+	if (str == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_STRING_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_STRING, &str);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static void state_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -290,6 +338,60 @@ static void strength_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
+static void roaming_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Roaming";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->roaming);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
+static void autoconnect_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "AutoConnect";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->autoconnect);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static void passphrase_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -341,6 +443,46 @@ static void passphrase_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
+static void apn_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	dbus_bool_t required;
+	const char *key = "SetupRequired";
+
+	if (service->path == NULL)
+		return;
+
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		return;
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		break;
+	}
+
+	required = (service->apn == NULL) ? TRUE : FALSE;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &required);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -395,6 +537,13 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	connman_dbus_dict_append_variant(&dict, "Favorite",
 					DBUS_TYPE_BOOLEAN, &service->favorite);
 
+	if (service->favorite == TRUE)
+		connman_dbus_dict_append_variant(&dict, "AutoConnect",
+				DBUS_TYPE_BOOLEAN, &service->autoconnect);
+	else
+		connman_dbus_dict_append_variant(&dict, "AutoConnect",
+					DBUS_TYPE_BOOLEAN, &service->favorite);
+
 	if (service->name != NULL)
 		connman_dbus_dict_append_variant(&dict, "Name",
 					DBUS_TYPE_STRING, &service->name);
@@ -404,7 +553,38 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		connman_dbus_dict_append_variant(&dict, "Roaming",
+					DBUS_TYPE_BOOLEAN, &service->roaming);
+
+		if (service->mcc != NULL && service->mnc != NULL) {
+			connman_dbus_dict_append_variant(&dict, "MCC",
+					DBUS_TYPE_STRING, &service->mcc);
+			connman_dbus_dict_append_variant(&dict, "MNC",
+					DBUS_TYPE_STRING, &service->mnc);
+		}
+
+		if (service->apn != NULL) {
+			connman_dbus_dict_append_variant(&dict, "APN",
+					DBUS_TYPE_STRING, &service->apn);
+
+			if (service->username != NULL)
+				connman_dbus_dict_append_variant(&dict,
+					"Username", DBUS_TYPE_STRING,
+							&service->username);
+
+			if (service->password != NULL)
+				connman_dbus_dict_append_variant(&dict,
+					"Password", DBUS_TYPE_STRING,
+							&service->password);
+
+			required = FALSE;
+		} else
+			required = TRUE;
+
+		connman_dbus_dict_append_variant(&dict, "SetupRequired",
+						DBUS_TYPE_BOOLEAN, &required);
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		if (service->passphrase != NULL &&
@@ -432,7 +612,9 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		break;
 	}
 
-	__connman_ipconfig_append_ipv4(service->ipconfig, &dict, "IPv4.");
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_append_ipv4(service->ipconfig,
+							&dict, "IPv4.");
 
 	dbus_message_iter_close_container(&array, &dict);
 
@@ -462,7 +644,26 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	type = dbus_message_iter_get_arg_type(&value);
 
-	if (g_str_equal(name, "Passphrase") == TRUE) {
+	if (g_str_has_prefix(name, "AutoConnect") == TRUE) {
+		connman_bool_t autoconnect;
+
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->favorite == FALSE)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &autoconnect);
+
+		if (service->autoconnect == autoconnect)
+			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+		service->autoconnect = autoconnect;
+
+		autoconnect_changed(service);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Passphrase") == TRUE) {
 		const char *passphrase;
 
 		if (type != DBUS_TYPE_STRING)
@@ -483,10 +684,72 @@ static DBusMessage *set_property(DBusConnection *conn,
 			connman_network_set_string(service->network,
 				"WiFi.Passphrase", service->passphrase);
 
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "APN") == TRUE) {
+		const char *apn;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &apn);
+
+		g_free(service->apn);
+		service->apn = g_strdup(apn);
+
+		apn_changed(service);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+						"Cellular.APN", service->apn);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Username") == TRUE) {
+		const char *username;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &username);
+
+		g_free(service->username);
+		service->username = g_strdup(username);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+					"Cellular.Username", service->username);
+
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Password") == TRUE) {
+		const char *password;
+
+		if (type != DBUS_TYPE_STRING)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &password);
+
+		g_free(service->password);
+		service->password = g_strdup(password);
+
+		if (service->network != NULL)
+			connman_network_set_string(service->network,
+					"Cellular.Password", service->password);
+
 		__connman_storage_save_service(service);
 	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
 		int err;
 
+		if (service->ipconfig == NULL)
+			return __connman_error_invalid_property(msg);
+
 		err = __connman_ipconfig_set_ipv4(service->ipconfig,
 							name + 5, &value);
 		if (err < 0)
@@ -557,6 +820,12 @@ static connman_bool_t is_connecting(struct connman_service *service)
 
 static connman_bool_t is_ignore(struct connman_service *service)
 {
+	if (service->autoconnect == FALSE)
+		return TRUE;
+
+	if (service->roaming == TRUE)
+		return TRUE;
+
 	if (service->ignore == TRUE)
 		return TRUE;
 
@@ -599,8 +868,10 @@ void __connman_service_auto_connect(void)
 		iter = g_sequence_iter_next(iter);
 	}
 
-	if (service != NULL)
+	if (service != NULL) {
+		service->userconnect = FALSE;
 		__connman_service_connect(service);
+	}
 }
 
 static void reply_pending(struct connman_service *service, int error)
@@ -630,35 +901,18 @@ static void reply_pending(struct connman_service *service, int error)
 static gboolean connect_timeout(gpointer user_data)
 {
 	struct connman_service *service = user_data;
-	connman_bool_t auto_connect = FALSE;
+	connman_bool_t autoconnect = FALSE;
 
 	DBG("service %p", service);
 
 	service->timeout = 0;
 
-	if (service->network != NULL) {
-		connman_bool_t connected;
-
-		connected = connman_network_get_connected(service->network);
-		if (connected == TRUE) {
-			__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_READY);
-			return FALSE;
-		}
-
+	if (service->network != NULL)
 		__connman_network_disconnect(service->network);
-	} else if (service->device != NULL) {
-		connman_bool_t disconnected;
-
-		disconnected = connman_device_get_disconnected(service->device);
-		if (disconnected == FALSE) {
-			__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_READY);
-			return FALSE;
-		}
-
+	else if (service->device != NULL)
 		__connman_device_disconnect(service->device);
-	}
+
+	__connman_ipconfig_disable(service->ipconfig);
 
 	if (service->pending != NULL) {
 		DBusMessage *reply;
@@ -670,12 +924,12 @@ static gboolean connect_timeout(gpointer user_data)
 		dbus_message_unref(service->pending);
 		service->pending = NULL;
 	} else
-		auto_connect = TRUE;
+		autoconnect = TRUE;
 
 	__connman_service_indicate_state(service,
 					CONNMAN_SERVICE_STATE_FAILURE);
 
-	if (auto_connect == TRUE)
+	if (autoconnect == TRUE && service->userconnect == FALSE)
 		__connman_service_auto_connect();
 
 	return FALSE;
@@ -707,10 +961,19 @@ static DBusMessage *connect_service(DBusConnection *conn,
 
 	service->ignore = FALSE;
 
+	service->userconnect = TRUE;
+	service->failcounter = 0;
+
 	service->pending = dbus_message_ref(msg);
 
 	err = __connman_service_connect(service);
 	if (err < 0) {
+		if (err == -ENOKEY) {
+			if (__connman_agent_request_passphrase(service,
+							NULL, NULL) == 0)
+				return NULL;
+		}
+
 		if (err != -EINPROGRESS) {
 			dbus_message_unref(service->pending);
 			service->pending = NULL;
@@ -735,6 +998,8 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 
 	DBG("service %p", service);
 
+	reply_pending(service, ECONNABORTED);
+
 	service->ignore = TRUE;
 
 	err = __connman_service_disconnect(service);
@@ -769,6 +1034,17 @@ static DBusMessage *remove_service(DBusConnection *conn,
 
 	passphrase_changed(service);
 
+	g_free(service->apn);
+	service->apn = NULL;
+
+	g_free(service->username);
+	service->username = NULL;
+
+	g_free(service->password);
+	service->password = NULL;
+
+	apn_changed(service);
+
 	connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
 
@@ -797,6 +1073,9 @@ static DBusMessage *move_before(DBusConnection *conn,
 
 	DBG("target %s", target->identifier);
 
+	if (target->state != service->state)
+		return __connman_error_invalid_service(msg);
+
 	g_get_current_time(&service->modified);
 	__connman_storage_save_service(service);
 
@@ -831,6 +1110,9 @@ static DBusMessage *move_after(DBusConnection *conn,
 
 	DBG("target %s", target->identifier);
 
+	if (target->state != service->state)
+		return __connman_error_invalid_service(msg);
+
 	g_get_current_time(&service->modified);
 	__connman_storage_save_service(service);
 
@@ -879,8 +1161,16 @@ static void service_free(gpointer user_data)
 	if (service->network != NULL)
 		connman_network_unref(service->network);
 
-	connman_ipconfig_unref(service->ipconfig);
+	if (service->ipconfig != NULL) {
+		connman_ipconfig_unref(service->ipconfig);
+		service->ipconfig = NULL;
+	}
 
+	g_free(service->mcc);
+	g_free(service->mnc);
+	g_free(service->apn);
+	g_free(service->username);
+	g_free(service->password);
 	g_free(service->profile);
 	g_free(service->name);
 	g_free(service->passphrase);
@@ -903,7 +1193,10 @@ void __connman_service_put(struct connman_service *service)
 
 		iter = g_hash_table_lookup(service_hash, service->identifier);
 		if (iter != NULL) {
+			reply_pending(service, ECONNABORTED);
+
 			__connman_service_disconnect(service);
+
 			g_sequence_remove(iter);
 		} else
 			service_free(service);
@@ -926,6 +1219,8 @@ static void __connman_service_initialize(struct connman_service *service)
 
 	service->ignore = FALSE;
 
+	service->userconnect = FALSE;
+
 	service->order = 0;
 }
 
@@ -948,15 +1243,6 @@ struct connman_service *connman_service_create(void)
 
 	__connman_service_initialize(service);
 
-	service->ipconfig = connman_ipconfig_create();
-	if (service->ipconfig == NULL) {
-		g_free(service);
-		return NULL;
-	}
-
-	connman_ipconfig_set_method(service->ipconfig,
-					CONNMAN_IPCONFIG_METHOD_DHCP);
-
 	return service;
 }
 
@@ -1009,6 +1295,20 @@ static gint service_compare(gconstpointer a, gconstpointer b,
 	if (service_a->favorite == FALSE && service_b->favorite == TRUE)
 		return 1;
 
+	if (service_a->type != service_b->type) {
+		switch (service_a->type) {
+		case CONNMAN_SERVICE_TYPE_UNKNOWN:
+		case CONNMAN_SERVICE_TYPE_ETHERNET:
+			break;
+		case CONNMAN_SERVICE_TYPE_WIFI:
+			return 1;
+		case CONNMAN_SERVICE_TYPE_WIMAX:
+		case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		case CONNMAN_SERVICE_TYPE_CELLULAR:
+			return -1;
+		}
+	}
+
 	return (gint) service_b->strength - (gint) service_a->strength;
 }
 
@@ -1090,6 +1390,8 @@ int __connman_service_indicate_state(struct connman_service *service,
 		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
 		state_changed(service);
 
+		reply_pending(service, ECONNABORTED);
+
 		__connman_service_disconnect(service);
 	}
 
@@ -1101,6 +1403,9 @@ int __connman_service_indicate_state(struct connman_service *service,
 
 		reply_pending(service, 0);
 
+		service->userconnect = FALSE;
+		service->failcounter = 0;
+
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
 
@@ -1114,8 +1419,19 @@ int __connman_service_indicate_state(struct connman_service *service,
 	}
 
 	if (state == CONNMAN_SERVICE_STATE_FAILURE) {
+		if (service->failcounter++ < MAX_CONNECT_RETRIES) {
+			connman_warn("Connecting again (try %d)",
+						service->failcounter);
+			__connman_service_disconnect(service);
+			__connman_service_connect(service);
+			return 0;
+		}
+
 		reply_pending(service, EIO);
 
+		if (service->userconnect == FALSE)
+			__connman_service_auto_connect();
+
 		g_get_current_time(&service->modified);
 		__connman_storage_save_service(service);
 	} else
@@ -1179,8 +1495,16 @@ static connman_bool_t prepare_network(struct connman_service *service)
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+		break;
 	case CONNMAN_NETWORK_TYPE_MBM:
 	case CONNMAN_NETWORK_TYPE_HSO:
+		connman_network_set_string(service->network,
+						"Cellular.APN", service->apn);
+
+		connman_network_set_string(service->network,
+					"Cellular.Username", service->username);
+		connman_network_set_string(service->network,
+					"Cellular.Password", service->password);
 		break;
 	}
 
@@ -1205,7 +1529,10 @@ int __connman_service_connect(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		if (service->apn == NULL)
+			return -EINVAL;
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		switch (service->security) {
@@ -1226,20 +1553,26 @@ int __connman_service_connect(struct connman_service *service)
 		if (prepare_network(service) == FALSE)
 			return -EINVAL;
 
+		__connman_ipconfig_enable(service->ipconfig);
+
 		err = __connman_network_connect(service->network);
 	} else if (service->device != NULL) {
 		if (service->favorite == FALSE)
 			return -ENOLINK;
 
+		__connman_ipconfig_enable(service->ipconfig);
+
 		err = __connman_device_connect(service->device);
 	} else
 		return -EOPNOTSUPP;
 
 	if (err < 0) {
-		if (err != -EINPROGRESS)
+		if (err != -EINPROGRESS) {
+			__connman_ipconfig_disable(service->ipconfig);
 			return err;
+		}
 
-		service->timeout = g_timeout_add_seconds(45,
+		service->timeout = g_timeout_add_seconds(CONNECT_TIMEOUT,
 						connect_timeout, service);
 
 		return -EINPROGRESS;
@@ -1254,8 +1587,6 @@ int __connman_service_disconnect(struct connman_service *service)
 
 	DBG("service %p", service);
 
-	reply_pending(service, ECONNABORTED);
-
 	if (service->network != NULL) {
 		err = __connman_network_disconnect(service->network);
 	} else if (service->device != NULL) {
@@ -1265,6 +1596,8 @@ int __connman_service_disconnect(struct connman_service *service)
 	} else
 		return -EOPNOTSUPP;
 
+	__connman_ipconfig_disable(service->ipconfig);
+
 	if (err < 0) {
 		if (err != -EINPROGRESS)
 			return err;
@@ -1457,6 +1790,8 @@ done:
 		service->passphrase = g_strdup(passphrase);
 	}
 
+	service->userconnect = TRUE;
+
 	err = __connman_service_connect(service);
 	if (err < 0 && err != -EINPROGRESS)
 		goto failed;
@@ -1551,6 +1886,65 @@ static int service_register(struct connman_service *service)
 	return 0;
 }
 
+static void service_up(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s up", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_down(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s down", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_lower_up(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s lower up", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_lower_down(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s lower down", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_ip_bound(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s ip bound", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_ip_release(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s ip release", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static const struct connman_ipconfig_ops service_ops = {
+	.up		= service_up,
+	.down		= service_down,
+	.lower_up	= service_lower_up,
+	.lower_down	= service_lower_down,
+	.ip_bound	= service_ip_bound,
+	.ip_release	= service_ip_release,
+};
+
+static void setup_ipconfig(struct connman_service *service, int index)
+{
+	if (index < 0)
+		return;
+
+	service->ipconfig = connman_ipconfig_create(index);
+	if (service->ipconfig == NULL)
+		return;
+
+	connman_ipconfig_set_method(service->ipconfig,
+					CONNMAN_IPCONFIG_METHOD_DHCP);
+
+	__connman_storage_load_service(service);
+
+	connman_ipconfig_set_data(service->ipconfig, service);
+
+	connman_ipconfig_set_ops(service->ipconfig, &service_ops);
+	connman_ipconfig_set_ops(service->ipconfig, NULL);
+}
+
 /**
  * __connman_service_lookup_from_device:
  * @device: device structure
@@ -1606,8 +2000,12 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 
 	service->type = __connman_device_get_service_type(device);
 
+	service->autoconnect = FALSE;
+
 	service->device = device;
 
+	setup_ipconfig(service, connman_device_get_index(device));
+
 	service_register(service);
 
 	__connman_profile_changed(TRUE);
@@ -1666,6 +2064,11 @@ struct connman_service *__connman_service_lookup_from_network(struct connman_net
 	return service;
 }
 
+const char *__connman_service_get_path(struct connman_service *service)
+{
+	return service->path;
+}
+
 unsigned int __connman_service_get_order(struct connman_service *service)
 {
 	GSequenceIter *iter;
@@ -1697,16 +2100,17 @@ static enum connman_service_type convert_network_type(struct connman_network *ne
 	switch (type) {
 	case CONNMAN_NETWORK_TYPE_UNKNOWN:
 	case CONNMAN_NETWORK_TYPE_VENDOR:
-	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
-	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
 		break;
-	case CONNMAN_NETWORK_TYPE_MBM:
-	case CONNMAN_NETWORK_TYPE_HSO:
-		return CONNMAN_SERVICE_TYPE_CELLULAR;
 	case CONNMAN_NETWORK_TYPE_WIFI:
 		return CONNMAN_SERVICE_TYPE_WIFI;
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 		return CONNMAN_SERVICE_TYPE_WIMAX;
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		return CONNMAN_SERVICE_TYPE_CELLULAR;
 	}
 
 	return CONNMAN_SERVICE_TYPE_UNKNOWN;
@@ -1740,6 +2144,24 @@ static enum connman_service_mode convert_wifi_security(const char *security)
 		return CONNMAN_SERVICE_SECURITY_UNKNOWN;
 }
 
+static enum connman_service_mode convert_cellular_mode(connman_uint8_t mode)
+{
+	switch (mode) {
+	case 0:
+	case 1:
+		return CONNMAN_SERVICE_MODE_GPRS;
+	case 3:
+		return CONNMAN_SERVICE_MODE_EDGE;
+	case 2:
+	case 4:
+	case 5:
+	case 6:
+		return CONNMAN_SERVICE_MODE_UMTS;
+	}
+
+	return CONNMAN_SERVICE_MODE_UNKNOWN;
+}
+
 static void update_from_network(struct connman_service *service,
 					struct connman_network *network)
 {
@@ -1765,6 +2187,15 @@ static void update_from_network(struct connman_service *service,
 	}
 
 	service->strength = connman_network_get_uint8(network, "Strength");
+	service->roaming = connman_network_get_bool(network, "Roaming");
+
+	if (service->strength == 0) {
+		/*
+		 * Filter out 0-values; it's unclear what they mean
+		 * and they cause anomalous sorting of the priority list.
+		 */
+		service->strength = strength;
+	}
 
 	str = connman_network_get_string(network, "WiFi.Mode");
 	service->mode = convert_wifi_mode(str);
@@ -1772,6 +2203,21 @@ static void update_from_network(struct connman_service *service,
 	str = connman_network_get_string(network, "WiFi.Security");
 	service->security = convert_wifi_security(str);
 
+	str = connman_network_get_string(network, "Cellular.MCC");
+	g_free(service->mcc);
+	service->mcc = g_strdup(str);
+
+	str = connman_network_get_string(network, "Cellular.MNC");
+	g_free(service->mnc);
+	service->mnc = g_strdup(str);
+
+	if (service->type == CONNMAN_SERVICE_TYPE_CELLULAR) {
+		connman_uint8_t value = connman_network_get_uint8(network,
+							"Cellular.Mode");
+
+		service->mode = convert_cellular_mode(value);
+	}
+
 	if (service->strength > strength && service->network != NULL) {
 		connman_network_unref(service->network);
 		service->network = connman_network_ref(network);
@@ -1826,10 +2272,25 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 
 	service->type = convert_network_type(network);
 
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		service->autoconnect = FALSE;
+		break;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		service->autoconnect = TRUE;
+		break;
+	}
+
 	service->state = CONNMAN_SERVICE_STATE_IDLE;
 
 	update_from_network(service, network);
 
+	setup_ipconfig(service, connman_network_get_index(network));
+
 	service_register(service);
 
 	__connman_profile_changed(TRUE);
@@ -1843,7 +2304,10 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 void __connman_service_update_from_network(struct connman_network *network)
 {
 	struct connman_service *service;
-	connman_uint8_t strength;
+	enum connman_service_mode mode;
+	connman_uint8_t strength, value;
+	connman_bool_t roaming;
+	GSequenceIter *iter;
 
 	service = __connman_service_lookup_from_network(network);
 	if (service == NULL)
@@ -1854,11 +2318,38 @@ void __connman_service_update_from_network(struct connman_network *network)
 
 	strength = connman_network_get_uint8(service->network, "Strength");
 	if (strength == service->strength)
-		return;
+		goto roaming;
 
 	service->strength = strength;
 
 	strength_changed(service);
+
+roaming:
+	roaming = connman_network_get_bool(service->network, "Roaming");
+	if (roaming == service->roaming)
+		goto done;
+
+	service->roaming = roaming;
+
+	roaming_changed(service);
+
+	iter = g_hash_table_lookup(service_hash, service->identifier);
+	if (iter != NULL)
+		g_sequence_sort_changed(iter, service_compare, NULL);
+
+done:
+	if (service->type != CONNMAN_SERVICE_TYPE_CELLULAR)
+		return;
+
+	value = connman_network_get_uint8(service->network, "Cellular.Mode");
+	mode = convert_cellular_mode(value);
+
+	if (mode == service->mode)
+		return;
+
+	service->mode = mode;
+
+	mode_changed(service);
 }
 
 void __connman_service_remove_from_network(struct connman_network *network)
@@ -1874,19 +2365,22 @@ void __connman_service_remove_from_network(struct connman_network *network)
 
 static int service_load(struct connman_service *service)
 {
+	const char *ident = service->profile;
 	GKeyFile *keyfile;
+	GError *error = NULL;
 	gchar *pathname, *data = NULL;
 	gsize length;
 	gchar *str;
+	connman_bool_t autoconnect;
 	unsigned int ssid_len;
 	int err = 0;
 
 	DBG("service %p", service);
 
-	if (service->profile == NULL)
+	if (ident == NULL)
 		return -EINVAL;
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, service->profile);
+	pathname = g_strdup_printf("%s/%s.profile", STORAGEDIR, ident);
 	if (pathname == NULL)
 		return -ENOMEM;
 
@@ -1964,9 +2458,24 @@ static int service_load(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		service->apn = g_key_file_get_string(keyfile,
+					service->identifier, "APN", NULL);
+
+		service->username = g_key_file_get_string(keyfile,
+					service->identifier, "Username", NULL);
+
+		service->username = g_key_file_get_string(keyfile,
+					service->identifier, "Password", NULL);
+
 		service->favorite = g_key_file_get_boolean(keyfile,
 				service->identifier, "Favorite", NULL);
 
+		autoconnect = g_key_file_get_boolean(keyfile,
+				service->identifier, "AutoConnect", &error);
+		if (error == NULL)
+			service->autoconnect = autoconnect;
+		g_clear_error(&error);
+
 		str = g_key_file_get_string(keyfile,
 				service->identifier, "Failure", NULL);
 		if (str != NULL) {
@@ -1990,7 +2499,8 @@ static int service_load(struct connman_service *service)
 		service->passphrase = str;
 	}
 
-	__connman_ipconfig_load(service->ipconfig, keyfile,
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_load(service->ipconfig, keyfile,
 					service->identifier, "IPv4.");
 
 done:
@@ -2001,6 +2511,7 @@ done:
 
 static int service_save(struct connman_service *service)
 {
+	const char *ident = service->profile;
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
@@ -2009,10 +2520,10 @@ static int service_save(struct connman_service *service)
 
 	DBG("service %p", service);
 
-	if (service->profile == NULL)
+	if (ident == NULL)
 		return -EINVAL;
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, service->profile);
+	pathname = g_strdup_printf("%s/%s.profile", STORAGEDIR, ident);
 	if (pathname == NULL)
 		return -ENOMEM;
 
@@ -2072,9 +2583,25 @@ update:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		if (service->apn != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+							"APN", service->apn);
+
+		if (service->username != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+						"Username", service->username);
+
+		if (service->password != NULL)
+			g_key_file_set_string(keyfile, service->identifier,
+						"Password", service->password);
+
 		g_key_file_set_boolean(keyfile, service->identifier,
 					"Favorite", service->favorite);
 
+		if (service->favorite == TRUE)
+			g_key_file_set_boolean(keyfile, service->identifier,
+					"AutoConnect", service->autoconnect);
+
 		if (service->state == CONNMAN_SERVICE_STATE_FAILURE) {
 			const char *failure = error2string(service->error);
 			if (failure != NULL)
@@ -2102,7 +2629,8 @@ update:
 		g_key_file_remove_key(keyfile, service->identifier,
 							"Passphrase", NULL);
 
-	__connman_ipconfig_save(service->ipconfig, keyfile,
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_save(service->ipconfig, keyfile,
 					service->identifier, "IPv4.");
 
 	data = g_key_file_to_data(keyfile, &length, NULL);
